#ifndef ITC_H
#define ITC_H

#include <iostream>
#include <string>
#include <queue>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/shared_ptr.hpp>

namespace itc {

typedef boost::function<void (void)> CallType;
	
struct ProcessingPolicy
{
	virtual void queue(CallType call) = 0;
	virtual void execute() = 0;
};

class WaitAndExecute : public ProcessingPolicy
{
public:
	WaitAndExecute();

	virtual void queue(CallType call);
	virtual void execute();

private:
	boost::mutex m_mutex;
	boost::condition_variable m_cond;
	bool m_req;
	std::queue<CallType>  m_callQ;	
};

class ProcessAndExecute : public ProcessingPolicy
{
public:
	ProcessAndExecute();

	virtual void queue(CallType call);
	virtual void execute();

private:
	virtual void process() = 0;

	boost::mutex m_mutex;
	std::queue<CallType>  m_callQ;
};

template <typename Callee, typename PolicyType = WaitAndExecute>
struct ITCThread
{
	ITCThread(const std::string name = "itcthr") :
		m_thr(), m_policy(new PolicyType), m_name(name), m_run(true)
	{
		m_thr = boost::thread(&ITCThread::run, this);
	}

	const std::string& getName()
	{
		return m_name;
	}

	void quit(int code)
	{
		doItPrivate(&ITCThread::doQuit, code);
	}

	void quit()
	{
		doItPrivate(&ITCThread::doQuit);
	}

	void join()
	{
		m_thr.join();
	}

	void doIt(CallType call)
	{
		m_policy->queue(call);
	}

	void doIt(void (Callee::*mf)(void))
	{
		doIt(CallType(boost::bind(mf, (Callee*)this)));
	}

	void doIt(void (*func)(void))
	{
		doIt(CallType(func));
	}

	template <typename ArgType>
	void doIt(void (Callee::*mf)(ArgType), ArgType arg)
	{
		doIt(CallType(boost::bind(mf, (Callee*)this, arg)));
	}

	template <typename ArgType>
	void doIt(void (*func)(ArgType), ArgType arg)
	{
		doIt(CallType(boost::bind(func, arg)));
	}

	template <typename ArgType1, typename ArgType2>
	void doIt(void (Callee::*mf)(ArgType1, ArgType2), ArgType1 arg1, ArgType2 arg2)
	{
		doIt(CallType(boost::bind(mf, (Callee*)this, arg1, arg2)));
	}

	template <typename ArgType1, typename ArgType2>
	void doIt(void (*func)(ArgType1, ArgType2), ArgType1 arg1, ArgType2 arg2)
	{
		doIt(CallType(boost::bind(func, arg1, arg2)));
	}

	template <typename ArgType1, typename ArgType2, typename ArgType3>
	void doIt(void (Callee::*mf)(ArgType1, ArgType2, ArgType3),
			  ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
	{
		doIt(CallType(boost::bind(mf, (Callee*)this, arg1, arg2, arg3)));
	}

	template <typename ArgType1, typename ArgType2, typename ArgType3>
	void doIt(void (*func)(ArgType1, ArgType2, ArgType3), 
			  ArgType1 arg1, ArgType2 arg2, ArgType3 arg3)
	{
		doIt(CallType(boost::bind(func, arg1, arg2, arg3)));
	}

private:
	void doItPrivate(void (ITCThread::*mf)(void))
	{
		doIt(CallType(boost::bind(mf, this)));
	}

	void doItPrivate(void (ITCThread::*mf)(int), int arg)
	{
		doIt(CallType(boost::bind(mf, this, arg)));
	}

	void run()
	{
		std::cout << m_name << ": starting..." << std::endl;
		while (m_run)
		{
			m_policy->execute();
		}
	}

	void doQuit(int code)
	{
		std::cout << m_name << ": exiting with code = " << code << std::endl;
		m_run = false;
	}

	void doQuit()
	{
		std::cout << m_name << ": exiting without code" << std::endl;
		m_run = false;
	}

	boost::thread m_thr;
	boost::shared_ptr<PolicyType> m_policy;
	std::string m_name;
	bool m_run;
};

template <typename Adaptee, typename PolicyType = WaitAndExecute>
struct ITCAdaptor : public Adaptee, public ITCThread<Adaptee, PolicyType>
{
	ITCAdaptor(const std::string& name) : ITCThread<Adaptee, PolicyType>(name) {}
};

struct Nothing
{
};

template <typename PolicyType = WaitAndExecute>
struct FunctionITCAdaptor : public ITCThread<Nothing, PolicyType>
{
	FunctionITCAdaptor(const std::string& name) :
		ITCThread<Nothing, PolicyType>(name) {}
};

};

#endif
